/*
 * phc -- the open source PHP compiler
 * See license/README.license for licensing information
 *
 * Main application module 
 */

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <dlfcn.h>
#include <unistd.h>
#include "generated/cmdline.h"
#include "process_ast/process_ast.h"
#include <phc/process_ast/PHP_unparser.h>
#include <phc/process_ast/XML_unparser.h>
#include <phc/process_ast/DOT_unparser.h>
#include <phc/process_ast/Generate_IR.h>
#include <phc/ast.h>
#include <phc/parse.h>

using namespace std;

struct gengetopt_args_info args_info;

void run_plugins(AST_php_script* php_script);

void sigsegv_handler(int)
{
	fprintf(stderr, "Segmentation fault\n");
	exit(-1);
}

void sigabrt_handler(int)
{
	exit(-1);
}

int main(int argc, char** argv)
{
	AST_php_script* php_script = NULL;

	// Synchronise C and C++ I/O
	ios_base::sync_with_stdio();

	// catch a seg fault - note this doesnt harm gdb, which will override this.
	// We do this so that tests will still continue if theres a seg fault
	signal(SIGSEGV, sigsegv_handler);
	signal(SIGABRT, sigabrt_handler);

	// Parse command line parameters 
	if(cmdline_parser(argc, argv, &args_info) != 0)
		exit(-1);

	if(args_info.inputs_num == 0)
	{
		php_script = parse(new String("-"), NULL, args_info.read_ast_xml_flag);
	}
	else
	{
		php_script = parse(new String(args_info.inputs[0]), NULL, args_info.read_ast_xml_flag);
		if (php_script == NULL)
		{
			error(ERR_FILE_NOT_FOUND, new String(args_info.inputs[0]), 0);
		}
	}

	// Make sure the inputs cannot be accessed globally
	args_info.inputs = NULL;

	if(php_script != NULL)
	{
		process_ast(php_script);

		/*
		 * Process shared objects (if any)
		 */

		run_plugins(php_script);

		if(args_info.generate_ir_flag)
		{
			Generate_IR generate_IR;
			php_script->visit(&generate_IR);
		}

		if(args_info.dump_php_flag)
		{
			PHP_unparser php_unparser;
			php_script->visit(&php_unparser);
		}

		if(args_info.dump_ast_xml_flag)
		{
			XML_unparser xml_unparser;
			php_script->visit(&xml_unparser);
		}

		if(args_info.dump_ast_dot_flag)
		{
			DOT_unparser dot_unparser;
			php_script->visit(&dot_unparser);
		}
	}

	return 0;
}
		
void run_plugins(AST_php_script* php_script)
{
	char in_cwd[256];
	char in_datadir[256];

	for(unsigned i = 0; i < args_info.run_given; i++)
	{
		void* handle; 
		const char* err;
		void (*past)(AST_php_script*);

		// Try opening the specified plugin from its default location, 
		// from the current working directory, and from DATADIR (in that order)
		handle = dlopen(args_info.run_arg[i], RTLD_LAZY);	
		err = dlerror();
		if(err)
		{
			// Try current directory
			getcwd(in_cwd, 256);
			strcat(in_cwd, "/");
			strcat(in_cwd, args_info.run_arg[i]);
			handle = dlopen(in_cwd, RTLD_LAZY);	
			err = dlerror();
		}
		if(err)
		{
			// Try DATADIR
			strcpy(in_datadir, DATADIR);
			strcat(in_datadir, "/" PACKAGE "/");
			strcat(in_datadir, args_info.run_arg[i]);
			handle = dlopen(in_datadir, RTLD_LAZY);	
			err = dlerror();
		}
		if(err)
		{
		 	error(ERR_PLUGIN_NOT_FOUND, NULL, 0, args_info.run_arg[i], in_cwd, in_datadir);
		}

		*(void **) (&past) = dlsym(handle, "process_ast");
		err = dlerror();
		if(err)
		{
			error(ERR_PLUGIN_INVALID, NULL, 0, args_info.run_arg[i]);
		}

		(*past)(php_script);
		
		dlclose(handle);
	}
}
