/*
 * phc -- the open source PHP compiler
 * See license/README.license for licensing information
 *
 * Compile PHP code to the phc intermediate representation
 */

#include "Generate_IR.h" 
#include <phc/lib/Integer.h>

int Generate_IR::get_temp(AST_node* node)
{
	Integer* temp = dynamic_cast<Integer*>(node->attrs->get("phc.temp"));
	return temp->value();
}

Generate_IR::Generate_IR()
{
	next_temp = 0;
	next_label = 0;
}

/*
 * Statements
 */

void Generate_IR::post_return(AST_return* in)
{
	if(in->expr == NULL)
		cout << "return NULL" << endl;
	else
		cout << "return ^t" << get_temp(in->expr) << endl;
}

void Generate_IR::pre_class_def(AST_class_def* in)
{
	current_class = in;
}

void Generate_IR::pre_method(AST_method* in)
{
	cout << endl << *current_class->class_name->value << "::";
	cout << *in->signature->method_name->value << ":" << endl;
	cout << "enter" << endl;
	next_temp = 0;
}

void Generate_IR::post_method(AST_method* in)
{
	cout << endl << "/* implicit return */" << endl;
	cout << "return NULL" << endl;
}

void Generate_IR::children_if(AST_if* in)
{
	int label = next_label;
	next_label += 2;

	cout << endl << "/* if condition */" << endl;
	in->expr->visit(this);
	cout << "t" << next_temp << " := ^t" << get_temp(in->expr) << ".cast_bool" << endl;
	next_temp++;
	cout << "t" << next_temp << " := call ^t" << next_temp - 1 << "()" << endl;
	cout << "if ^t" << next_temp << " goto l" << label << endl;
	next_temp++;

	cout << endl << "/* false branch */" << endl;
	in->iffalse->visit(this);
	cout << "goto l" << label + 1 << endl;

	cout << endl << "/* true branch */" << endl;
	cout << "l" << label << ":" << endl;
	in->iftrue->visit(this);

	cout << endl << "/* after if */" << endl;
	cout << "l" << label + 1 << ":" << endl;
}

void Generate_IR::children_while(AST_while* in)
{
	int label = next_label;
	next_label += 2;
	
	continue_stack.push_back(label);
	
	cout << endl << "/* while condition */" << endl;
	cout << "l" << label << ":" << endl;
	in->expr->visit(this);
	cout << "t" << next_temp << " := ^t" << get_temp(in->expr) << ".cast_bool" << endl;
	next_temp++;
	cout << "t" << next_temp << " := call ^t" << next_temp - 1 << "()" << endl;
	next_temp++;
	cout << "t" << next_temp << " := ^t" << next_temp - 1 << ".op!" << endl;
	next_temp++;
	cout << "t" << next_temp << " := call ^t" << next_temp - 1 << "()" << endl;
	cout << "if ^t" << next_temp << " goto l" << label + 1 << endl;
	next_temp++;	

	cout << endl << "/* while body */" << endl;
	in->statements->visit(this);
	cout << "goto l" << label << endl;

	cout << endl << "/* after while */" << endl;
	cout << "l" << label + 1 << ":" << endl;
	
	continue_stack.pop_back();	
}

void Generate_IR::children_for(AST_for* in)
{
	int label = next_label;
	next_label += 2;
	
	cout << endl << "/* for-loop initialisation */" << endl;
	in->init->visit(this);

	cout << endl << "/* for-loop condition */" << endl;
	cout << "l" << label << ":" << endl;
	in->cond->visit(this);

	continue_stack.push_back(label);

	cout << "t" << next_temp << " := ^t" << get_temp(in->cond) << ".cast_bool" << endl;
	next_temp++;
	cout << "t" << next_temp << " := call ^t" << next_temp - 1 << "()" << endl;
	next_temp++;
	cout << "t" << next_temp << " := ^t" << next_temp - 1 << ".op!" << endl;
	next_temp++;
	cout << "t" << next_temp << " := call ^t" << next_temp - 1 << "()" << endl;
	cout << "if ^t" << next_temp << " goto l" << label + 1 << endl;
	next_temp++;

	cout << endl << "/* for-loop body */" << endl;
	in->statements->visit(this);

	cout << endl << "/* for-loop increment */" << endl;
	in->incr->visit(this);
	cout << "goto l" << label << endl;

	cout << endl << "/* after for-loop */" << endl;
	cout << "l" << label + 1 << ":" << endl;

	continue_stack.pop_back();
}

void Generate_IR::children_continue(AST_continue* in)
{
	if(in->expr)
	{
		Token_int* expr = dynamic_cast<Token_int*>(in->expr);
		
		if(!expr)
		{
			cout << "Error: arbitrary expressions not supported in 'continue' on line " << in->get_line_number() << endl;
			exit(-1);
		}
		
		if(continue_stack.size() < (unsigned)expr->value)
		{
			cout << "Error: invalid argument to 'continue' on  line " << in->get_line_number() << endl;
			exit(-1);
		}
	
		list<int>::const_iterator i = continue_stack.end();
		for(int j = 0; j < expr->value; j++, i--);
		
		cout << "goto l" << *i << endl;
	}
	else
	{
		cout << "goto l" << continue_stack.back() << endl;
	}
	
}

/*
 * Expressions
 */

void Generate_IR::post_method_invocation(AST_method_invocation* in)
{
	// TODO: Deal with non-static methods
	Token_class_name* cn = new Token_class_name(WILDCARD);
	Token_method_name* mn = new Token_method_name(WILDCARD);
	if(in->target->match(cn) && in->method_name->match(mn))
	{
		AST_actual_parameter_list::const_iterator i;	
		cout << "t" << next_temp << " := call " << *cn->value << "::" << *mn->value << "(";
		for(i = in->actual_parameters->begin(); i != in->actual_parameters->end(); i++)
		{
			if(i != in->actual_parameters->begin()) cout << ", ";
			// TODO: deal with is_ref
			cout << "^t" << get_temp((*i)->expr);
		}
		cout << ")" << endl;
		in->attrs->set("phc.temp", new Integer(next_temp));
		next_temp++;
	}
}

void Generate_IR::post_assignment(AST_assignment* in)
{
	if(in->is_ref)
	{
		cout << "^t" << get_temp(in->variable) << " := ^t" << get_temp(in->expr) << endl;
	}
	else
	{
		cout << "t" << next_temp << " := ^t" << get_temp(in->expr) << ".clone" << endl;
		next_temp++;
		cout << "t" << next_temp << " := call ^t" << next_temp - 1 << "()" << endl;
		cout << "^t" << get_temp(in->variable) << " := ^t" << next_temp << endl;
		next_temp++;
	}
}

void Generate_IR::post_variable(AST_variable* in)
{
	// TODO: deals with the simple case only 
	Token_variable_name* name = new Token_variable_name(WILDCARD);
	AST_variable* pattern = new AST_variable(NULL, name, new AST_expr_list, NULL);

	if(in->match(pattern))
	{
		cout << "t" << next_temp << " := " << *name->value << endl;
		in->attrs->set("phc.temp", new Integer(next_temp++));
	}
}

void Generate_IR::children_conditional_expr(AST_conditional_expr* in)
{
	int label = next_label;
	next_label += 2;
	int result = next_temp++;
	
	cout << endl << "/* ?: condition */" << endl;
	in->cond->visit(this);
	cout << "t" << next_temp << " := ^t" << get_temp(in->cond) << ".cast_bool" << endl;
	next_temp++;
	cout << "t" << next_temp << " := call ^t" << next_temp - 1 << "()" << endl;
	cout << "if ^t" << next_temp << " goto l" << label << endl;
	next_temp++;

	cout << endl << "/* ? branch */" << endl;
	in->iffalse->visit(this);
	cout << "t" << result << " := ^t" << get_temp(in->iffalse) << endl;
	cout << "goto l" << label + 1 << endl;

	cout << endl << "/* : branch */" << endl;
	cout << "l" << label << ":" << endl;
	in->iftrue->visit(this);
	cout << "t" << result << " := ^t" << get_temp(in->iftrue) << endl;

	cout << endl << "/* after ?: */" << endl;
	cout << "l" << label + 1 << ":" << endl;

	in->attrs->set("phc.temp", new Integer(result));
}

void Generate_IR::children_bin_op(AST_bin_op* in)
{
	if(*in->op->value == "&&" || *in->op->value == "and")
	{
		int label = next_label++;
		int result = next_temp++;
		
		cout << endl << "/* left of && */" << endl;
		in->left->visit(this);
		cout << "t" << result << " := ^t" << get_temp(in->left) << endl;
		cout << "t" << next_temp << " := ^t" << result << ".cast_bool" << endl;
		next_temp++;
		cout << "t" << next_temp << " := call ^t" << next_temp -1 << "()" << endl;
		next_temp++;
		cout << "t" << next_temp << " := ^t" << next_temp - 1 << ".op!" << endl;
		next_temp++;
		cout << "t" << next_temp << " := call ^t" << next_temp - 1 << "()" << endl;
		cout << "if ^t" << next_temp << " goto l" << label << endl;
		next_temp++;

		cout << endl << "/* right of && */" << endl;
		in->right->visit(this);
		cout << "t" << result << " := ^t" << get_temp(in->right) << endl;
		
		cout << endl << "/* after && */" << endl;
		cout << "l" << label << ":" << endl;

		in->attrs->set("phc.temp", new Integer(result));
	}
	else if(*in->op->value == "||" || *in->op->value == "or")
	{
		int label = next_label++;
		int result = next_temp++;
		
		cout << endl << "/* left of || */" << endl;
		in->left->visit(this);
		cout << "t" << result << " := ^t" << get_temp(in->left) << endl;
		cout << "t" << next_temp << " := ^t" << result << ".cast_bool" << endl;
		next_temp++;
		cout << "t" << next_temp << " := call ^t" << next_temp -1 << "()" << endl;
		cout << "if ^t" << next_temp << " goto l" << label << endl;
		next_temp++;

		cout << endl << "/* right of || */" << endl;
		in->right->visit(this);
		cout << "t" << result << " := ^t" << get_temp(in->right) << endl;
		
		cout << endl << "/* after || */" << endl;
		cout << "l" << label << ":" << endl;

		in->attrs->set("phc.temp", new Integer(result));
	}
	else
	{
		in->left->visit(this);
		in->right->visit(this);

		cout << "t" << next_temp << " := ^t" << get_temp(in->left) << ".op" << *in->op->value << endl; 
		next_temp++;
		cout << "t" << next_temp << " := call ^t" << next_temp - 1 << "(^t" << get_temp(in->right) << ")" << endl;
		in->attrs->set("phc.temp", new Integer(next_temp++));
	}
}

void Generate_IR::post_unary_op(AST_unary_op* in)
{
	cout << "t" << next_temp << " := ^t" << get_temp(in->expr) << ".op" << *in->op->value << endl;
	next_temp++;
	cout << "t" << next_temp << " := call ^t" << next_temp - 1 << "()" << endl;
	in->attrs->set("phc.temp", new Integer(next_temp++));
}

void Generate_IR::post_literal(AST_literal* in)
{
	cout << "t" << next_temp << " := #" << *in->get_value_as_string() << endl;
	in->attrs->set("phc.temp", new Integer(next_temp++));
}
